/*
    CASD-0: main.m
    Copyright (C) 2004  Macaulay Institute

    This file is part of CASD (Casuistry And Social Dilemmas), an agent-based
    model designed to study the behaviour of case-based reasoners when they
    face social dilemmas. Social dilemmas are modelled as n-player games.

    CASD-0 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    CASD-0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Luis R. Izquierdo,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      l.izquierdo@macaulay.ac.uk
*/

/*

CASD - Casuistry And Social Dilemmas

main.m

*/

#import "ObserverSwarm.h"
#import "BatchSwarm.h"
#import "Verbosity.h"
#import <errno.h>
#import <simtools.h>
#import <simtoolsgui.h>
#import <stdlib.h>
#import <string.h>
#import <sys/types.h>
#import <sys/stat.h>
#import <random.h>

#define DEFAULT_VERBOSITY_FILE "casd.verby"

int main(int argc, const char **argv) {
  id theTopLevelSwarm;
  const char *params;
  const char **fakeargv;
  const char *verbosityFile = DEFAULT_VERBOSITY_FILE;
  const char *deterministicStr = NULL;
  int fakeargc;
  int i, j;
  int swarmArgStart, paramArg;
  int verbosity = 0;
  int stopPeriod = 0;
  BOOL stopPeriodSpecified = NO;
  int checkTime = 0;
  BOOL checkTimeSpecified = NO;
  BOOL blackAndWhite = NO;
  BOOL deterministic = NO;
  struct stat fstat;
  BOOL showParseArgs = NO;

  /* We've not used the Swarm argument handler, so we have to handle our
     own arguments. These use a + instead of a - to indicate they are 
     options, and currently there are six options available:
     +v and +V determine verbosity -- the level of messages to print to 
     stdout about the progress of the model, +t determines the termination
     time of the simulation, +c is to check for cyclic behaviour, +b is to 
     show graphs in black & white, and +d is to do deterministic runs, as 
     opposed to runs with stochastic components.
  
     +v <arg> takes an integer argument specifying the level of verbosity 
     (higher numbers mean more messages, according to the verbosity file) 
     +V <arg> takes a filename argument -- the file referred to containing 
     a list of message types that are generated by the system, and a 
     verbosity level at or above which the messages are to be printed. This
     file is called the 'verbosity file'.
     +t <arg> takes an integer argument specifying the stop timestep. The 
     simulation will stop at timestep t. When it stops, it requires the 
     user to press Start in the control panel to go on.
     +c <arg> takes an integer argument specifying the timestep at which the 
     the program will start to search for a cyclic behaviour (check time). 
     At check time, if there is a cyclic behaviour and the level of verbosity
     is high enough (not less than the verbosity required for 
     CyclicBehaviour), the cycle will be detected. Stable cooperation
     and stable defection are detected always as soon as they occur, if the 
     level of verbosity is high enough (not less than the verbosity required
     for Equilibria). If the simulation has not reached a cycle at check 
     time (either because check time is too low or because the initial 
     conditions do not lead to a cyclic behaviour), the program issues a 
     warning as soon as it detects that situation, and stops the simulation.
     +b makes graphs in black & white.
     +d <arg> takes a string argument such as CCDDCDDCD. Each letter 
     represents a decision (Cooperate or Defect). When, according to the 
     program requirement specifications, a random decision should take place,
     if this option is used, the 'random' decision will be taken from the 
     string argument from left to right. As an example, if the string 
     argument is CDD, the first 'random' decision will be Cooperate, the 
     second one Defect, and the third one Defect. Note that this is for the 
     model, not for a specific Agent. If the string is too short, the 
     program aborts, but it kindly lets you know what length would be 
     sufficient, according to your parameters.
     +w shows the arguments passed.
     
     The + options must appear first after the command.
  */

  if(argc < 2) {

    fprintf(stderr, "\nUsage: %s "
	    "\n\t[ +v <verbosityLevel> ]"
	    "\n\t[ +V <verbosityFile> ]"
	    "\n\t[ +t <stopTime> ]"
	    "\n\t[ +c <checkTime> ]"
	    "\n\t[ +b ]"
	    "\n\t[ +d <decisionsString> ]"
	    "\n\t[ +w ]"
	    "\n\tparameterFile"
	    "\n\t[Swarm arguments]  (e.g. [ -s ]  [ -S <seed> ]  [ -b ])\n"
	    , argv[0]);
    exit(1);
  }

  printf("%s -- Flags:\n", argv[0]);
  for(i = 1; i < argc && argv[i][0] == '+'; i++) {
				// Loop from the first argument to the program
				// until the end of the arguments or the end
				// of arguments beginning +
    printf("%d: %s", i, argv[i]);
    switch(argv[i][1]) {
    case 'v':
      i++;
      if(argc <= i) {
        // raiseEvent(InvalidArgument, "+v flag takes an argument");
	fprintf(stderr, "+v flag takes an argument\n");
	abort();
      }
      printf(" %s", argv[i]);
      verbosity = atoi(argv[i]);
      break;
    case 'V':
      i++;
      if(argc <= i) {
	// raiseEvent(InvalidArgument, "+V flag takes an argument");
	fprintf(stderr, "+V flag takes an argument\n");
	abort();
      }
      printf(" %s", argv[i]);
      verbosityFile = argv[i];
      break;
    case 't':
      i++;
      if(argc <= i) {
        // raiseEvent(InvalidArgument, "+t flag takes an argument");
	fprintf(stderr, "+t flag takes an argument\n");
	abort();
      }
      printf(" %s", argv[i]);
      stopPeriod = atoi(argv[i]);
      stopPeriodSpecified = YES;
      break;
    case 'c':
      i++;
      if(argc <= i) {
        // raiseEvent(InvalidArgument, "+c flag takes an argument");
	fprintf(stderr, "+c flag takes an argument\n");
	abort();
      }
      printf(" %s", argv[i]);
      checkTime = atoi(argv[i]);
      checkTimeSpecified = YES;
      break;
    case 'b':
      blackAndWhite = YES;
      break;
    case 'd':
      i++;
      if(argc <= i) {
        printf("+d flag takes an argument");
	fprintf(stderr, "+d flag takes an argument\n");
	abort();
      }
      printf(" %s", argv[i]);
      deterministicStr = argv[i];
      deterministic = YES;
      break;
    case 'w':
      showParseArgs = YES;
      break;
    default:
      // raiseEvent(InvalidArgument, "%s flag not recognised", argv[i]);
      fprintf(stderr, "%s flag not recognised\n", argv[i]);
      abort();
    }
    printf("\n");
  }
  paramArg = i;			// Remember the last argument we were about
				// to use for this-program-specific arguments
				// -- this is the parameter file argument
  swarmArgStart = paramArg + 1;
				// The next argument will be an argument
				// for swarm-specific arguments
 
  if(showParseArgs) {
				// Display a message about parsing the
				// arguments to the program if the verbosity
				// level has been set appropriately
    printf("Parameter file at argument %d, Swarm args start at %d\n",
	   paramArg, swarmArgStart);
  }

  // The last argument before any swarm flags should be the parameter file

  if(argc <= paramArg) {
    // raiseEvent(InvalidArgument, "You must specify the parameter file "
    //           "(after any CASD specific flags and before "
    //	         "any Swarm specific flags)");
    fprintf(stderr, "You must specify the parameter file "
	    "(after any CASD specific flags and before "
	    "any Swarm specific flags)\n");
    abort();
  }

  // Now we create a fake argument vector to send to the Swarm initialisation

  params = argv[paramArg];
  if(stat(params, &fstat) != 0) {
    fprintf(stderr, "There is a problem with the parameter file:\n");
    perror(params);
    abort();
  }

  fakeargv = malloc((argc - swarmArgStart + 2) * sizeof(char *));
  if(fakeargv == NULL) {
    // raiseEvent(OutOfMemory, "Could not allocate memory: %s", 
    // strerror(errno));
    fprintf(stderr, "Could not allocate memory: %s\n", strerror(errno));
    abort();
  }
  fakeargc = 0;
  for(i = 0, j = 0; i < argc; i++) {
    if(showParseArgs) printf("Analysing argument %s\n", argv[i]);
    if(i > 0 && i < swarmArgStart) {
      if(showParseArgs) {
	printf("Argument %s is not for Swarm\n", argv[i]);
      }
      continue;
    }
    if(showParseArgs) {
      printf("Argument %s is for Swarm\n", argv[i]);
    }
    fakeargv[j] = argv[i];
    fakeargc++;
    j++;
  }

  /////////////////////////////////////////////////////////////////////////////

  // Initialise Swarm

  initSwarm(fakeargc, fakeargv);
  
  // No methods should be called before the call to initSwarm. This
  // upsets Swarm-2.2.

  /////////////////////////////////////////////////////////////////////////////

  printf("Seed: %u\n", [randomGenerator getInitialSeed]);

  //Initialise Verbosity

  if(strcmp(verbosityFile, DEFAULT_VERBOSITY_FILE) != 0
     || stat(verbosityFile, &fstat) == 0) {
    [Verbosity initialise: verbosityFile];
				// Initialise the verbosity class if a
				// verbosity file has been specified using +V
				// or the default verbosity file exists.
				// N.B. If this file does not exist, the
				// default is not to show anything regardless
				// of the verbosity level.
  }
  [Verbosity setLevel: verbosity];
				// Set the verbosity level 

  // Start the model running

  if (swarmGUIMode == 1) {
    // We've got graphics, so make a full ObserverSwarm to get GUI objects
    if([Verbosity showProgress]) printf("Building observer\n");
    theTopLevelSwarm = [ObserverSwarm createBegin: globalZone];
    [theTopLevelSwarm setParameterFile: params];
    theTopLevelSwarm = [theTopLevelSwarm createEnd];
    [theTopLevelSwarm setBlackAndWhiteRasters: blackAndWhite];
  }
  else {
    // No graphics - make a batchmode swarm and run it.
    if([Verbosity showProgress]) printf("Building batchSwarm\n");
    theTopLevelSwarm = [BatchSwarm createBegin: globalZone];
    [theTopLevelSwarm setParameterFile: params];
    theTopLevelSwarm = [theTopLevelSwarm createEnd];    
  }

  if(stopPeriodSpecified) [theTopLevelSwarm setStopPeriod: stopPeriod];
  if(checkTimeSpecified) [theTopLevelSwarm setCheckTime: checkTime];
  if(deterministic) [theTopLevelSwarm setDeterministicStr: deterministicStr];
  [theTopLevelSwarm buildObjects];
  [theTopLevelSwarm buildActions];
  [theTopLevelSwarm activateIn: nil];
  if([Verbosity showProgress]) printf("Starting the simulation\n");
  [theTopLevelSwarm go];

  free(fakeargv);
  return 0;
}
